# dptdatasource.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide DPT recordsets combined with bsddb3 style cursor access.

DPTDataSource is a container for the recordset.

CursorRS provides a bsddb3 style cursor for the recordset.

"""

from dptdb import dptapi

from basesup.dptbase import CursorDPT

from ..core.dataclient import DataSource


class DPTDataSource(DataSource):
    
    """Provide bsddb3 style cursor access to recordset of arbitrary records.
    """

    def __init__(self, dbhome, dbset, dbname, newrow=None):
        """Define bsddb3 style cursor interface on a recordset.
        
        See superclass for description of arguments.

        """
        super(DPTDataSource, self).__init__(
            dbhome, dbset, dbname, newrow=newrow)

        self.recordset = None
        self._fieldvalue = dptapi.APIFieldValue()
        self.dbhome.get_dptfiles()[self.dbset]._sources[self] = None
        
    def close(self):
        if self.recordset is not None:
            try:
                del self.dbhome.get_dptfiles()[self.dbset]._sources[self]
            except:
                pass
            self.dbhome.get_dptfiles()[self.dbset].get_database(
                ).DestroyRecordSet(self.recordset)
            self.recordset = None

    def get_cursor(self):
        """Return cursor on record set, or list, associated with datasource."""
        if self.recordset:
            c = CursorRS(
                self.dbhome.get_dptfiles()[self.dbset],
                self.dbhome.get_dptfiles()[self.dbname]._primary,
                recordset=self.recordset)
        else:
            c = CursorRS(
                self.dbhome.get_dptfiles()[self.dbset],
                self.dbhome.get_dptfiles()[self.dbname]._primary,
                recordset=self.dbhome.get_database(
                    self.dbset, self.dbname).CreateRecordList())
        if c:
            self.dbhome.get_dptfiles()[self.dbset]._clientcursors[c] = True
        return c

    def join_field_occurrences(self, record, field):
        """Return concatenated occurrences of field."""
        i = 1
        v = []
        while record.GetFieldValue(field, self._fieldvalue, i):
            v.append(self._fieldvalue.ExtractString())
            i += 1
        return ''.join(v)

    def set_recordset(self, records):
        """Set self.recordset to records (a DPT record set or list)."""
        if self.recordset:
            self.dbhome.get_database(
                self.dbset, self.dbname).DestroyRecordSet(self.recordset)
        self.recordset = records


class CursorRS(CursorDPT):
    
    """Provide a bsddb3 style cursor for a recordset of arbitrary records.

    The cursor does not support partial keys because the records in the
    recordset do not have an implied order (apart from the accidential order
    of existence on the database).

    """

    def set_partial_key(self, partial):
        """Set partial key to None.  Always.
        
        Always set to None because the record set or list should be trimmed
        to the required records before passing to the cursor.
        
        """
        # See comments in _CursorDPT class definition for reasons why _partial
        # is now constrained to be None always. Originally a design choice.
        super(CursorRS, self).set_partial_key(None)

    def get_partial(self):
        """Return self._partial"""
        return self._partial

